@@ -40,7 +40,7 @@ class :xhpy:html-element(:x:primitive):
     buf = '<' + self.tagName
     attributes = self.getAttributes()
     for key in attributes:
-      val = attributes[key]
+      val = unicode(attributes[key])
       if val is not None and val is not False:
         buf += ' ' + htmlspecialchars(key) + '="' +\
                      htmlspecialchars(val, True) + '"'
@@ -456,10 +456,32 @@ class :input(:xhpy:html-singleton):
     string alt, bool checked, bool disabled, int maxlength, string name,
     bool readonly, int size, string src,
     enum {
-      "button", "checkbox", "file", "hidden", "image", "password", "radio",
-      "reset", "submit", "text"
+      "button",
+      "checkbox",
+      "color",
+      "date",
+      "datetime",
+      "datetime-local",
+      "email",
+      "file",
+      "hidden",
+      "image",
+      "month",
+      "number",
+      "password",
+      "radio",
+      "range",
+      "reset",
+      "search",
+      "submit",
+      "tel",
+      "text",
+      "time",
+      "url",
+      "week"
     } type,
-    string value
+    string value,
+    string required
   category %flow, %phrase, %interactive
   def __init__(self, attributes={}, children=[]):
     super(:xhpy:html-element, self).__init__(attributes, children)
@@ -534,7 +556,8 @@ class :meta(:xhpy:html-singleton):
     enum {
       "content-type", "content-style-type", "expires", "refresh", "set-cookie"
     } http-equiv,
-    string http-equiv, string name, string scheme
+    string http-equiv, string name, string scheme,
+    string charset
   category %metadata
   def __init__(self, attributes={}, children=[]):
     super(:xhpy:html-element, self).__init__(attributes, children)
@@ -760,7 +783,9 @@ class :td(:xhpy:html-element):
 
 
 class :textarea(:xhpy:pseudo-singleton):
-  attribute int cols, int rows, bool disabled, string name, bool readonly
+  attribute
+    int cols, int rows, bool disabled, string name, bool readonly,
+    string placeholder, string required
   category %flow, %phrase, %interactive
   def __init__(self, attributes={}, children=[]):
     super(:xhpy:html-element, self).__init__(attributes, children)
